#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2016 Christ Stefan <contact@stefanchrist.eu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys
import os
from argparse import ArgumentParser
from subprocess import Popen, PIPE, DEVNULL


sys_encoding = sys.getfilesystemencoding()


def parseRFC822(output):
    # Example
    #    TENC:
    #    WXXX: ():
    #    PRIV:  (unimplemented)
    #    PRIV:  (unimplemented)
    #    APIC: (picture)[, 3]: image/jpeg, 17274 bytes
    #    TCON: Punk Rock (121)
    #    TIT3: Kopfüber in die Hölle
    #    TRCK: 16
    #    TALB: Die Bestie in Menschengestalt
    #    PRIV:  (unimplemented)
    #    PRIV:  (unimplemented)
    #    TOPE:
    #    TPE1: Die Ärzte
    #    COMM: ()[]:
    #    COMM: (Media Jukebox: Tool Name)[]: Media Jukebox
    #    COMM: (Media Jukebox: Tool Version)[]: 12.0.49
    #
    # Handle invalid RFC822 outptu from id3v2: Missing indention!
    #
    #    Filename: fixture-invalid-RFC822/Narcotic.mp3
    #    USLT: ()[deu]: So you face it with a smileThere is no need to cryFor a
    #    trifle's more than thisWill you still recall my nameAnd the
    #    ...
    #    SeekLyrics
    #    APIC: ()[, 0]: image/jpeg, 6094 bytes
    #    TIT2: Narcotic
    #

    data = {}
    for line in output.split(b"\n"):
        if line == b"":
            continue

        if b": " not in line:
            # Don't raise an exception. Invalid RFC822 output.  Unfolding
            # currently no implemented!
            continue

        tag, value = line.split(b": ", 1)
        if value != b" (unimplemented)":
            # Ignore  unknown data
            data[tag] = value

    return data


class NoMetaDataInMP3(Exception):
    pass


# Function converts the file format specific encoding to a python string object
# that can represent all characters.
def getNewNameId3v2(path):
    p = Popen([b"id3v2", b"-R", path],
              stdout=PIPE,
              stderr=DEVNULL)
    stdout, _ = p.communicate()

    if p.returncode != 0:
        raise RuntimeError("Program id3v2 fails for file %s with error code %d!" % (path, p.returncode))

    values = parseRFC822(stdout)

    if b"TPE1" not in values:
        raise NoMetaDataInMP3("No valid artist in file %s!" % (path,))
    if b"TALB" not in values:
        raise NoMetaDataInMP3("No valid album in file %s!" % (path,))
    if b"TIT2" not in values:
        raise NoMetaDataInMP3("No track in file file %s!" % (path,))

    # id3v2 honors environment variable LANG.
    # TODO catch not encodable exception
    artist = values[b"TPE1"].decode(sys_encoding)
    album = values[b"TALB"].decode(sys_encoding)
    track = values[b"TIT2"].decode(sys_encoding)

    return artist, album, track


# Specification of ID3v1 says strings are encoding as 'ISO-8859-1'.
# Non-European languages are not supported.  See wiki for ID3v1.
#
# Function converts the file format specific encoding to a python string
# object that can represent all characters.
def getNewNameMp3Info(path):
    p = Popen([b"mp3info", b"-p", b"%a\n%l\n%t", path],
              stdout=PIPE,
              stderr=DEVNULL)
    e = p.wait()
    stdout = p.stdout.read()
    p.stdout.close()

    if e != 0:
        raise RuntimeError("Program mp3info fails for file %s with error code %d!" % (path, e))

    artist, album, track = stdout.split(b"\n")
    if len(artist) == 0 or len(album) == 0 or len(track) == 0:
        raise NoMetaDataInMP3("No valid artist, album and track information in file %s!"
                              % (path,))

    # TODO catch not encodable exception
    artist = artist.decode("ISO-8859-1")
    album = album.decode("ISO-8859-1")
    track = track.decode("ISO-8859-1")

    return artist, album, track


def replaceEvilCharactersInBytes(s):
    f = lambda c: 0 < c < 128 and c or ord("?")
    return b''.join(f(c) for c in s)


def replaceEvilCharactersInString(s):
    f = lambda c: 0 < ord(c) < 128 and c or '?'
    return ''.join(f(c) for c in s)


def moveFile(path, new_path):
    assert isinstance(path, bytes)
    assert isinstance(new_path, bytes)
    os.makedirs(os.path.dirname(new_path), exist_ok=True)

    # NOTE Check first and then move is racy!
    if os.path.exists(new_path):
        print("Warning: Destination file %s already exists!" % (new_path,),
              file=sys.stderr)
    else:
        os.rename(path, new_path)


def moveFileAndTryToStrip(input_file, output_path, new_filename):
    assert isinstance(new_filename, bytes)

    new_path = os.path.join(output_path, new_filename)
    try:
        moveFile(input_file, new_path)
    except OSError as e:
        print("Error while moving file %s: %s" % (input_file, e),
              file=sys.stderr)
        # Mostly Error:
        #    [Errno 84] Invalid or incomplete multibyte or wide character: b'MP3/The Lord Of The Rings/Die R\xfcckkehr Des K\xf6nigs'
        # Try to fix it with a stripped version of the new filename.
        new_filename_stripped = replaceEvilCharactersInBytes(new_filename)
        print("Stripping filename %s -> %s" % (new_filename, new_filename_stripped),
              file=sys.stderr)
        new_path_stripped = os.path.join(output_path, new_filename_stripped)
        try:
            moveFile(input_file, new_path_stripped)
        except OSError as e:
            print("Error while moving file %s: %s" % (input_file, e), file=sys.stderr)


def osWalkAllFiles(input_path):
    for dirpath, dirnames, files in os.walk(input_path):
        for filename in files:
            yield os.path.join(dirpath, filename)


# Encode a string object as bytes of encoding sys.getfilesystemencoding().  If
# one character is not printable in the system encoding, replace _all_
# non-ASCII bytes with '?'.
def encodeAndReplace(s):
    # Convert string to suitable bytes name in the system. Use
    # sys.getfilesystemencoding().
    try:
        b = s.encode(sys_encoding)
    except UnicodeEncodeError:
        # Is raised by print, if a character cannot be converted to the
        # sys.getfilesystemencoding() on output.
        s = replaceEvilCharactersInString(s)
        b = s.encode(sys_encoding)
    return b


def main():
    parser = ArgumentParser()
    parser.add_argument(dest="input", type=str,
                        help="Directory of input MP3 files.")
    parser.add_argument(dest="output", type=str,
                        help="Directory of sorted output files.")
    parser.add_argument("--dry-run", dest="dryrun",
                        action="store_true", default=False,
                        help="Don't copy files, just show move operations.")

    args = parser.parse_args()

    dryrun = args.dryrun
    input_path = args.input.encode(sys_encoding)
    output_path = args.output.encode(sys_encoding)

    # Get all input files
    files_mp3 = (path for path in osWalkAllFiles(input_path)
                 if os.path.isfile(path) and path.endswith(b".mp3"))

    # Move files
    for input_file in files_mp3:
        try:
            try:
                artist, album, track = getNewNameId3v2(input_file)
            except NoMetaDataInMP3 as e:
                artist, album, track = getNewNameMp3Info(input_file)

            new_filename = os.path.join(artist, album, track + ".mp3")
        except (NoMetaDataInMP3, RuntimeError) as e:
            print("# Ignore file %s. Error: %s" % (input_file, e))
            continue

        assert isinstance(input_file, bytes)
        assert isinstance(output_path, bytes)
        assert isinstance(new_filename, str)

        new_filename = encodeAndReplace(new_filename)

        if not dryrun:
            moveFileAndTryToStrip(input_file, output_path, new_filename)
        else:
            # TODO make valid bash/shell output
            sys.stdout.buffer.write(b"mv " + input_file + b" " + new_filename + b"\n")

    return 0


if __name__ == "__main__":
    sys.exit(main())
